@@ -1,60 +1,60 @@
-/*  This file is part of the Vc library.
+/*  This file is part of the Vc library. {{{
+Copyright © 2009-2015 Matthias Kretz <kretz@kde.org>
 
-    Copyright (C) 2009-2012 Matthias Kretz <kretz@kde.org>
+Redistribution and use in source and binary forms, with or without
+modification, are permitted provided that the following conditions are met:
+    * Redistributions of source code must retain the above copyright
+      notice, this list of conditions and the following disclaimer.
+    * Redistributions in binary form must reproduce the above copyright
+      notice, this list of conditions and the following disclaimer in the
+      documentation and/or other materials provided with the distribution.
+    * Neither the names of contributing organizations nor the
+      names of its contributors may be used to endorse or promote products
+      derived from this software without specific prior written permission.
 
-    Vc is free software: you can redistribute it and/or modify
-    it under the terms of the GNU Lesser General Public License as
-    published by the Free Software Foundation, either version 3 of
-    the License, or (at your option) any later version.
+THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
+ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
+DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
+(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
+ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
+(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
+SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 
-    Vc is distributed in the hope that it will be useful, but
-    WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU Lesser General Public License for more details.
+}}}*/
 
-    You should have received a copy of the GNU Lesser General Public
-    License along with Vc.  If not, see <http://www.gnu.org/licenses/>.
+#ifndef VC_SSE_INTRINSICS_H_
+#define VC_SSE_INTRINSICS_H_
 
-*/
-
-#ifndef SSE_INTRINSICS_H
-#define SSE_INTRINSICS_H
-
-#include "../common/windows_fix_intrin.h"
-
-// The GCC xxxintrin.h headers do not make sure that the intrinsics have C linkage. This not really
-// a problem, unless there is another place where the exact same functions are declared. Then the
-// linkage must be the same, otherwise it won't compile. Such a case occurs on Windows, where the
-// intrin.h header (included indirectly via unistd.h) declares many SSE intrinsics again.
-extern "C" {
-// MMX
-#include <mmintrin.h>
-// SSE
-#include <xmmintrin.h>
-// SSE2
-#include <emmintrin.h>
-}
-
-#include "../common/fix_clang_emmintrin.h"
-
-#include "const_data.h"
-#include <cstdlib>
-#include "macros.h"
-
-#ifdef __3dNOW__
-extern "C" {
-#include <mm3dnow.h>
-}
+#ifdef Vc_MSVC
+#include <intrin.h>
+#else
+#include <x86intrin.h>
 #endif
 
-/*OUTER_NAMESPACE_BEGIN*/
-namespace Vc
-{
-namespace SSE
-{
-    enum VectorAlignmentEnum { VectorAlignment = 16 };
+#include "../common/storage.h"
+#include "const_data.h"
+#include <cstdlib>
+#include "types.h"
+#include "debug.h"
 
-#if defined(VC_GCC) && VC_GCC < 0x40600 && !defined(VC_DONT_FIX_SSE_SHIFT)
+#if defined(Vc_GCC) && !defined(__OPTIMIZE__)
+// GCC uses lots of old-style-casts in macros that disguise as intrinsics
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wold-style-cast"
+#endif
+
+namespace Vc_VERSIONED_NAMESPACE
+{
+namespace SseIntrinsics
+{
+    using SSE::c_general;
+
+    constexpr std::size_t VectorAlignment = 16;
+
+#if defined(Vc_GCC) && Vc_GCC < 0x40600 && !defined(Vc_DONT_FIX_SSE_SHIFT)
     static Vc_INTRINSIC Vc_CONST __m128i _mm_sll_epi16(__m128i a, __m128i count) { __asm__("psllw %1,%0" : "+x"(a) : "x"(count)); return a; }
     static Vc_INTRINSIC Vc_CONST __m128i _mm_sll_epi32(__m128i a, __m128i count) { __asm__("pslld %1,%0" : "+x"(a) : "x"(count)); return a; }
     static Vc_INTRINSIC Vc_CONST __m128i _mm_sll_epi64(__m128i a, __m128i count) { __asm__("psllq %1,%0" : "+x"(a) : "x"(count)); return a; }
@@ -63,7 +63,7 @@ namespace SSE
     static Vc_INTRINSIC Vc_CONST __m128i _mm_srl_epi64(__m128i a, __m128i count) { __asm__("psrlq %1,%0" : "+x"(a) : "x"(count)); return a; }
 #endif
 
-#ifdef VC_GCC
+#ifdef Vc_GCC
     // Redefine the mul/add/sub intrinsics to use GCC-specific operators instead of builtin
     // functions. This way the fp-contraction optimization step kicks in and creates FMAs! :)
     static Vc_INTRINSIC Vc_CONST __m128d _mm_mul_pd(__m128d a, __m128d b) { return static_cast<__m128d>(static_cast<__v2df>(a) * static_cast<__v2df>(b)); }
@@ -74,14 +74,9 @@ namespace SSE
     static Vc_INTRINSIC Vc_CONST __m128  _mm_sub_ps(__m128  a, __m128  b) { return static_cast<__m128 >(static_cast<__v4sf>(a) - static_cast<__v4sf>(b)); }
 #endif
 
-#if defined(VC_GNU_ASM) && !defined(NVALGRIND)
-    static Vc_INTRINSIC __m128i Vc_CONST _mm_setallone() { __m128i r; __asm__("pcmpeqb %0,%0":"=x"(r)); return r; }
-#else
-    static Vc_INTRINSIC __m128i Vc_CONST _mm_setallone() { __m128i r = _mm_setzero_si128(); return _mm_cmpeq_epi8(r, r); }
-#endif
-    static Vc_INTRINSIC __m128i Vc_CONST _mm_setallone_si128() { return _mm_setallone(); }
-    static Vc_INTRINSIC __m128d Vc_CONST _mm_setallone_pd() { return _mm_castsi128_pd(_mm_setallone()); }
-    static Vc_INTRINSIC __m128  Vc_CONST _mm_setallone_ps() { return _mm_castsi128_ps(_mm_setallone()); }
+    static Vc_INTRINSIC Vc_CONST __m128i _mm_setallone_si128() { return _mm_load_si128(reinterpret_cast<const __m128i *>(Common::AllBitsSet)); }
+    static Vc_INTRINSIC Vc_CONST __m128d _mm_setallone_pd() { return _mm_load_pd(reinterpret_cast<const double *>(Common::AllBitsSet)); }
+    static Vc_INTRINSIC Vc_CONST __m128  _mm_setallone_ps() { return _mm_load_ps(reinterpret_cast<const float *>(Common::AllBitsSet)); }
 
     static Vc_INTRINSIC __m128i Vc_CONST _mm_setone_epi8 ()  { return _mm_set1_epi8(1); }
     static Vc_INTRINSIC __m128i Vc_CONST _mm_setone_epu8 ()  { return _mm_setone_epi8(); }
@@ -98,67 +93,108 @@ namespace SSE
     static Vc_INTRINSIC __m128d Vc_CONST _mm_setsignmask_pd(){ return _mm_load_pd(reinterpret_cast<const double *>(c_general::signMaskDouble)); }
     static Vc_INTRINSIC __m128  Vc_CONST _mm_setsignmask_ps(){ return _mm_load_ps(reinterpret_cast<const float *>(c_general::signMaskFloat)); }
 
-    //X         static Vc_INTRINSIC __m128i Vc_CONST _mm_setmin_epi8 () { return _mm_slli_epi8 (_mm_setallone_si128(),  7); }
-    static Vc_INTRINSIC __m128i Vc_CONST _mm_setmin_epi16() { return _mm_load_si128(reinterpret_cast<const __m128i *>(c_general::minShort)); }
-    static Vc_INTRINSIC __m128i Vc_CONST _mm_setmin_epi32() { return _mm_load_si128(reinterpret_cast<const __m128i *>(c_general::signMaskFloat)); }
+    static Vc_INTRINSIC __m128i Vc_CONST setmin_epi8 () { return _mm_set1_epi8(-0x80); }
+    static Vc_INTRINSIC __m128i Vc_CONST setmin_epi16() { return _mm_load_si128(reinterpret_cast<const __m128i *>(c_general::minShort)); }
+    static Vc_INTRINSIC __m128i Vc_CONST setmin_epi32() { return _mm_load_si128(reinterpret_cast<const __m128i *>(c_general::signMaskFloat)); }
+    static Vc_INTRINSIC __m128i Vc_CONST setmin_epi64() { return _mm_load_si128(reinterpret_cast<const __m128i *>(c_general::signMaskDouble)); }
 
-    //X         static Vc_INTRINSIC __m128i Vc_CONST _mm_cmplt_epu8 (__m128i a, __m128i b) { return _mm_cmplt_epi8 (
-    //X                 _mm_xor_si128(a, _mm_setmin_epi8 ()), _mm_xor_si128(b, _mm_setmin_epi8 ())); }
-    //X         static Vc_INTRINSIC __m128i Vc_CONST _mm_cmpgt_epu8 (__m128i a, __m128i b) { return _mm_cmpgt_epi8 (
-    //X                 _mm_xor_si128(a, _mm_setmin_epi8 ()), _mm_xor_si128(b, _mm_setmin_epi8 ())); }
-    static Vc_INTRINSIC __m128i Vc_CONST _mm_cmplt_epu16(__m128i a, __m128i b) { return _mm_cmplt_epi16(
-            _mm_xor_si128(a, _mm_setmin_epi16()), _mm_xor_si128(b, _mm_setmin_epi16())); }
-    static Vc_INTRINSIC __m128i Vc_CONST _mm_cmpgt_epu16(__m128i a, __m128i b) { return _mm_cmpgt_epi16(
-            _mm_xor_si128(a, _mm_setmin_epi16()), _mm_xor_si128(b, _mm_setmin_epi16())); }
-    static Vc_INTRINSIC __m128i Vc_CONST _mm_cmplt_epu32(__m128i a, __m128i b) { return _mm_cmplt_epi32(
-            _mm_xor_si128(a, _mm_setmin_epi32()), _mm_xor_si128(b, _mm_setmin_epi32())); }
-    static Vc_INTRINSIC __m128i Vc_CONST _mm_cmpgt_epu32(__m128i a, __m128i b) { return _mm_cmpgt_epi32(
-            _mm_xor_si128(a, _mm_setmin_epi32()), _mm_xor_si128(b, _mm_setmin_epi32())); }
-} // namespace SSE
-} // namespace Vc
-/*OUTER_NAMESPACE_END*/
-
-// SSE3
-#ifdef VC_IMPL_SSE3
-extern "C" {
-#include <pmmintrin.h>
-}
-#endif
-// SSSE3
-#ifdef VC_IMPL_SSSE3
-extern "C" {
-#include <tmmintrin.h>
-}
-#define mm_abs_epi8  _mm_abs_epi8
-#define mm_abs_epi16 _mm_abs_epi16
-#define mm_abs_epi32 _mm_abs_epi32
-#define mm_alignr_epi8 _mm_alignr_epi8
-/*OUTER_NAMESPACE_BEGIN*/
-namespace Vc
-{
-namespace SSE
-{
-
-    // not overriding _mm_set1_epi8 because this one should only be used for non-constants
-    static Vc_INTRINSIC __m128i Vc_CONST set1_epi8(int a) {
-#if defined(VC_GCC) && VC_GCC < 0x40500
-        return _mm_shuffle_epi8(_mm_cvtsi32_si128(a), _mm_setzero_si128());
+#if defined(Vc_IMPL_XOP)
+    static Vc_INTRINSIC __m128i Vc_CONST cmplt_epu8(__m128i a, __m128i b) { return _mm_comlt_epu8(a, b); }
+    static Vc_INTRINSIC __m128i Vc_CONST cmpgt_epu8(__m128i a, __m128i b) { return _mm_comgt_epu8(a, b); }
+    static Vc_INTRINSIC __m128i Vc_CONST cmplt_epu16(__m128i a, __m128i b) { return _mm_comlt_epu16(a, b); }
+    static Vc_INTRINSIC __m128i Vc_CONST cmpgt_epu16(__m128i a, __m128i b) { return _mm_comgt_epu16(a, b); }
+    static Vc_INTRINSIC __m128i Vc_CONST cmplt_epu32(__m128i a, __m128i b) { return _mm_comlt_epu32(a, b); }
+    static Vc_INTRINSIC __m128i Vc_CONST cmpgt_epu32(__m128i a, __m128i b) { return _mm_comgt_epu32(a, b); }
+    static Vc_INTRINSIC __m128i Vc_CONST cmplt_epu64(__m128i a, __m128i b) { return _mm_comlt_epu64(a, b); }
+    static Vc_INTRINSIC __m128i Vc_CONST cmpgt_epu64(__m128i a, __m128i b) { return _mm_comgt_epu64(a, b); }
 #else
-        // GCC 4.5 nows about the pshufb improvement
-        return _mm_set1_epi8(a);
+    static Vc_INTRINSIC __m128i Vc_CONST cmplt_epu8(__m128i a, __m128i b)
+    {
+        return _mm_cmplt_epi8(_mm_xor_si128(a, setmin_epi8()),
+                              _mm_xor_si128(b, setmin_epi8()));
+    }
+    static Vc_INTRINSIC __m128i Vc_CONST cmpgt_epu8(__m128i a, __m128i b)
+    {
+        return _mm_cmpgt_epi8(_mm_xor_si128(a, setmin_epi8()),
+                              _mm_xor_si128(b, setmin_epi8()));
+    }
+    static Vc_INTRINSIC __m128i Vc_CONST cmplt_epu16(__m128i a, __m128i b)
+    {
+        return _mm_cmplt_epi16(_mm_xor_si128(a, setmin_epi16()),
+                               _mm_xor_si128(b, setmin_epi16()));
+    }
+    static Vc_INTRINSIC __m128i Vc_CONST cmpgt_epu16(__m128i a, __m128i b)
+    {
+        return _mm_cmpgt_epi16(_mm_xor_si128(a, setmin_epi16()),
+                               _mm_xor_si128(b, setmin_epi16()));
+    }
+    static Vc_INTRINSIC __m128i Vc_CONST cmplt_epu32(__m128i a, __m128i b)
+    {
+        return _mm_cmplt_epi32(_mm_xor_si128(a, setmin_epi32()),
+                               _mm_xor_si128(b, setmin_epi32()));
+    }
+    static Vc_INTRINSIC __m128i Vc_CONST cmpgt_epu32(__m128i a, __m128i b)
+    {
+        return _mm_cmpgt_epi32(_mm_xor_si128(a, setmin_epi32()),
+                               _mm_xor_si128(b, setmin_epi32()));
+    }
+    Vc_INTRINSIC __m128i Vc_CONST cmpgt_epi64(__m128i a, __m128i b)
+    {
+#ifdef Vc_IMPL_SSE4_2
+        return _mm_cmpgt_epi64(a, b);
+#else
+        const auto aa = _mm_xor_si128(a, _mm_srli_epi64(setmin_epi32(),32));
+        const auto bb = _mm_xor_si128(b, _mm_srli_epi64(setmin_epi32(),32));
+        const auto gt = _mm_cmpgt_epi32(aa, bb);
+        const auto eq = _mm_cmpeq_epi32(aa, bb);
+        // Algorithm:
+        // 1. if the high 32 bits of gt are true, make the full 64 bits true
+        // 2. if the high 32 bits of gt are false and the high 32 bits of eq are true,
+        //    duplicate the low 32 bits of gt to the high 32 bits (note that this requires
+        //    unsigned compare on the lower 32 bits, which is the reason for the xors
+        //    above)
+        // 3. else make the full 64 bits false
+
+        const auto gt2 =
+            _mm_shuffle_epi32(gt, 0xf5);  // dup the high 32 bits to the low 32 bits
+        const auto lo =
+            _mm_shuffle_epi32(_mm_and_si128(_mm_srli_epi64(eq, 32), gt), 0xa0);
+        return _mm_or_si128(gt2, lo);
 #endif
     }
+    static Vc_INTRINSIC __m128i Vc_CONST cmpgt_epu64(__m128i a, __m128i b)
+    {
+        return cmpgt_epi64(_mm_xor_si128(a, setmin_epi64()),
+                           _mm_xor_si128(b, setmin_epi64()));
+    }
+#endif
+}  // namespace SseIntrinsics
+}  // namespace Vc
+
+// SSSE3
+#ifdef Vc_IMPL_SSSE3
+namespace Vc_VERSIONED_NAMESPACE
+{
+namespace SseIntrinsics
+{
+    // not overriding _mm_set1_epi8 because this one should only be used for non-constants
+    Vc_INTRINSIC Vc_CONST __m128i abs_epi8(__m128i a) { return _mm_abs_epi8(a); }
+    Vc_INTRINSIC Vc_CONST __m128i abs_epi16(__m128i a) { return _mm_abs_epi16(a); }
+    Vc_INTRINSIC Vc_CONST __m128i abs_epi32(__m128i a) { return _mm_abs_epi32(a); }
+    template <int s> Vc_INTRINSIC Vc_CONST __m128i alignr_epi8(__m128i a, __m128i b)
+    {
+        return _mm_alignr_epi8(a, b, s & 0x1fu);
+    }
+}  // namespace SseIntrinsics
+}  // namespace Vc
 
-} // namespace SSE
-} // namespace Vc
-/*OUTER_NAMESPACE_END*/
 #else
-/*OUTER_NAMESPACE_BEGIN*/
-namespace Vc
+
+namespace Vc_VERSIONED_NAMESPACE
 {
-namespace SSE
+namespace SseIntrinsics
 {
-    static Vc_INTRINSIC __m128i Vc_CONST mm_abs_epi8 (__m128i a) {
+    Vc_INTRINSIC Vc_CONST __m128i abs_epi8 (__m128i a) {
         __m128i negative = _mm_cmplt_epi8 (a, _mm_setzero_si128());
         return _mm_add_epi8 (_mm_xor_si128(a, negative), _mm_and_si128(negative,  _mm_setone_epi8()));
     }
@@ -172,19 +208,17 @@ namespace SSE
     //   a xor -1 -> -a - 1
     //   -1 >> 31 -> 1
     //   -a - 1 + 1 -> -a
-    static Vc_INTRINSIC __m128i Vc_CONST mm_abs_epi16(__m128i a) {
+    Vc_INTRINSIC Vc_CONST __m128i abs_epi16(__m128i a) {
         __m128i negative = _mm_cmplt_epi16(a, _mm_setzero_si128());
         return _mm_add_epi16(_mm_xor_si128(a, negative), _mm_srli_epi16(negative, 15));
     }
-    static Vc_INTRINSIC __m128i Vc_CONST mm_abs_epi32(__m128i a) {
+    Vc_INTRINSIC Vc_CONST __m128i abs_epi32(__m128i a) {
         __m128i negative = _mm_cmplt_epi32(a, _mm_setzero_si128());
         return _mm_add_epi32(_mm_xor_si128(a, negative), _mm_srli_epi32(negative, 31));
     }
-    static Vc_INTRINSIC __m128i Vc_CONST set1_epi8(int a) {
-        return _mm_set1_epi8(a);
-    }
-    static Vc_INTRINSIC __m128i Vc_CONST mm_alignr_epi8(__m128i a, __m128i b, const int s) {
-        switch (s) {
+    template <int s> Vc_INTRINSIC Vc_CONST __m128i alignr_epi8(__m128i a, __m128i b)
+    {
+        switch (s & 0x1fu) {
             case  0: return b;
             case  1: return _mm_or_si128(_mm_slli_si128(a, 15), _mm_srli_si128(b,  1));
             case  2: return _mm_or_si128(_mm_slli_si128(a, 14), _mm_srli_si128(b,  2));
@@ -220,72 +254,143 @@ namespace SSE
         }
         return _mm_setzero_si128();
     }
-
-} // namespace SSE
-} // namespace Vc
-/*OUTER_NAMESPACE_END*/
-
+}  // namespace SseIntrinsics
+}  // namespace Vc
 #endif
 
 // SSE4.1
-#ifdef VC_IMPL_SSE4_1
-extern "C" {
-#include <smmintrin.h>
+#ifdef Vc_IMPL_SSE4_1
+namespace Vc_VERSIONED_NAMESPACE
+{
+namespace SseIntrinsics
+{
+Vc_INTRINSIC Vc_CONST __m128i cmpeq_epi64(__m128i a, __m128i b)
+{
+    return _mm_cmpeq_epi64(a, b);
 }
-/*OUTER_NAMESPACE_BEGIN*/
-namespace Vc
+template <int index> Vc_INTRINSIC Vc_CONST int extract_epi32(__m128i v)
 {
-namespace SSE
+    return _mm_extract_epi32(v, index);
+}
+Vc_INTRINSIC Vc_CONST __m128d blendv_pd(__m128d a, __m128d b, __m128d c)
 {
-#define mm_blendv_pd _mm_blendv_pd
-#define mm_blendv_ps _mm_blendv_ps
-#define mm_blendv_epi8 _mm_blendv_epi8
-#define mm_blend_epi16 _mm_blend_epi16
-#define mm_blend_ps _mm_blend_ps
-#define mm_blend_pd _mm_blend_pd
-
-#define mm_min_epi32 _mm_min_epi32
-#define mm_max_epi32 _mm_max_epi32
-#define mm_min_epu32 _mm_min_epu32
-#define mm_max_epu32 _mm_max_epu32
-//#define mm_min_epi16 _mm_min_epi16
-//#define mm_max_epi16 _mm_max_epi16
-#define mm_min_epu16 _mm_min_epu16
-#define mm_max_epu16 _mm_max_epu16
-#define mm_min_epi8  _mm_min_epi8
-#define mm_max_epi8  _mm_max_epi8
-
-#define mm_cvtepu16_epi32 _mm_cvtepu16_epi32
-#define mm_cvtepu8_epi16 _mm_cvtepu8_epi16
-#define mm_cvtepi8_epi16 _mm_cvtepi8_epi16
-#define mm_cvtepu16_epi32 _mm_cvtepu16_epi32
-#define mm_cvtepi16_epi32 _mm_cvtepi16_epi32
-#define mm_cvtepu8_epi32 _mm_cvtepu8_epi32
-#define mm_cvtepi8_epi32 _mm_cvtepi8_epi32
-#define mm_stream_load_si128 _mm_stream_load_si128
-// TODO
-} // namespace SSE
-} // namespace Vc
-/*OUTER_NAMESPACE_END*/
+    return _mm_blendv_pd(a, b, c);
+}
+Vc_INTRINSIC Vc_CONST __m128 blendv_ps(__m128 a, __m128 b, __m128 c)
+{
+    return _mm_blendv_ps(a, b, c);
+}
+Vc_INTRINSIC Vc_CONST __m128i blendv_epi8(__m128i a, __m128i b, __m128i c)
+{
+    return _mm_blendv_epi8(a, b, c);
+}
+template <int mask> Vc_INTRINSIC Vc_CONST __m128d blend_pd(__m128d a, __m128d b)
+{
+    return _mm_blend_pd(a, b, mask);
+}
+template <int mask> Vc_INTRINSIC Vc_CONST __m128 blend_ps(__m128 a, __m128 b)
+{
+    return _mm_blend_ps(a, b, mask);
+}
+template <int mask> Vc_INTRINSIC Vc_CONST __m128i blend_epi16(__m128i a, __m128i b)
+{
+    return _mm_blend_epi16(a, b, mask);
+}
+Vc_INTRINSIC Vc_CONST __m128i max_epi8(__m128i a, __m128i b)
+{
+    return _mm_max_epi8(a, b);
+}
+Vc_INTRINSIC Vc_CONST __m128i max_epi32(__m128i a, __m128i b)
+{
+    return _mm_max_epi32(a, b);
+}
+Vc_INTRINSIC Vc_CONST __m128i max_epu16(__m128i a, __m128i b)
+{
+    return _mm_max_epu16(a, b);
+}
+Vc_INTRINSIC Vc_CONST __m128i max_epu32(__m128i a, __m128i b)
+{
+    return _mm_max_epu32(a, b);
+}
+Vc_INTRINSIC Vc_CONST __m128i min_epu16(__m128i a, __m128i b)
+{
+    return _mm_min_epu16(a, b);
+}
+Vc_INTRINSIC Vc_CONST __m128i min_epu32(__m128i a, __m128i b)
+{
+    return _mm_min_epu32(a, b);
+}
+Vc_INTRINSIC Vc_CONST __m128i min_epi8(__m128i a, __m128i b)
+{
+    return _mm_min_epi8(a, b);
+}
+Vc_INTRINSIC Vc_CONST __m128i min_epi32(__m128i a, __m128i b)
+{
+    return _mm_min_epi32(a, b);
+}
+Vc_INTRINSIC Vc_CONST __m128i cvtepu8_epi16(__m128i epu8)
+{
+    return _mm_cvtepu8_epi16(epu8);
+}
+Vc_INTRINSIC Vc_CONST __m128i cvtepi8_epi16(__m128i epi8)
+{
+    return _mm_cvtepi8_epi16(epi8);
+}
+Vc_INTRINSIC Vc_CONST __m128i cvtepu16_epi32(__m128i epu16)
+{
+    return _mm_cvtepu16_epi32(epu16);
+}
+Vc_INTRINSIC Vc_CONST __m128i cvtepi16_epi32(__m128i epu16)
+{
+    return _mm_cvtepi16_epi32(epu16);
+}
+Vc_INTRINSIC Vc_CONST __m128i cvtepu8_epi32(__m128i epu8)
+{
+    return _mm_cvtepu8_epi32(epu8);
+}
+Vc_INTRINSIC Vc_CONST __m128i cvtepi8_epi32(__m128i epi8)
+{
+    return _mm_cvtepi8_epi32(epi8);
+}
+Vc_INTRINSIC Vc_PURE __m128i stream_load_si128(__m128i *mem)
+{
+    return _mm_stream_load_si128(mem);
+}
+}  // namespace SseIntrinsics
+}  // namespace Vc
 #else
-/*OUTER_NAMESPACE_BEGIN*/
-namespace Vc
+
+namespace Vc_VERSIONED_NAMESPACE
 {
-namespace SSE
+namespace SseIntrinsics
 {
-    static Vc_INTRINSIC __m128d mm_blendv_pd(__m128d a, __m128d b, __m128d c) {
+    Vc_INTRINSIC Vc_CONST __m128i cmpeq_epi64(__m128i a, __m128i b) {
+        auto tmp = _mm_cmpeq_epi32(a, b);
+        return _mm_and_si128(tmp, _mm_shuffle_epi32(tmp, 1*1 + 0*4 + 3*16 + 2*64));
+    }
+    template <int index> Vc_INTRINSIC Vc_CONST int extract_epi32(__m128i v)
+    {
+#ifdef Vc_USE_BUILTIN_VECTOR_TYPES
+        typedef int int32v4 __attribute__((__vector_size__(16)));
+        return reinterpret_cast<const MayAlias<int32v4> &>(v)[index];
+#else
+        return _mm_cvtsi128_si32(_mm_srli_si128(v, index * 4));
+#endif
+    }
+    Vc_INTRINSIC Vc_CONST __m128d blendv_pd(__m128d a, __m128d b, __m128d c) {
         return _mm_or_pd(_mm_andnot_pd(c, a), _mm_and_pd(c, b));
     }
-    static Vc_INTRINSIC __m128  mm_blendv_ps(__m128  a, __m128  b, __m128  c) {
+    Vc_INTRINSIC Vc_CONST __m128  blendv_ps(__m128  a, __m128  b, __m128  c) {
         return _mm_or_ps(_mm_andnot_ps(c, a), _mm_and_ps(c, b));
     }
-    static Vc_INTRINSIC __m128i mm_blendv_epi8(__m128i a, __m128i b, __m128i c) {
+    Vc_INTRINSIC Vc_CONST __m128i blendv_epi8(__m128i a, __m128i b, __m128i c) {
         return _mm_or_si128(_mm_andnot_si128(c, a), _mm_and_si128(c, b));
     }
 
     // only use the following blend functions with immediates as mask and, of course, compiling
     // with optimization
-    static Vc_INTRINSIC __m128d mm_blend_pd(__m128d a, __m128d b, const int mask) {
+    template <int mask> Vc_INTRINSIC Vc_CONST __m128d blend_pd(__m128d a, __m128d b)
+    {
         switch (mask) {
         case 0x0:
             return a;
@@ -300,7 +405,8 @@ namespace SSE
             return a; // should never be reached, but MSVC needs it else it warns about 'not all control paths return a value'
         }
     }
-    static Vc_INTRINSIC __m128  mm_blend_ps(__m128  a, __m128  b, const int mask) {
+    template <int mask> Vc_INTRINSIC Vc_CONST __m128 blend_ps(__m128 a, __m128 b)
+    {
         __m128i c;
         switch (mask) {
         case 0x0:
@@ -357,7 +463,8 @@ namespace SSE
         __m128 _c = _mm_castsi128_ps(c);
         return _mm_or_ps(_mm_andnot_ps(_c, a), _mm_and_ps(_c, b));
     }
-    static Vc_INTRINSIC __m128i mm_blend_epi16(__m128i a, __m128i b, const int mask) {
+    template <int mask> Vc_INTRINSIC Vc_CONST __m128i blend_epi16(__m128i a, __m128i b)
+    {
         __m128i c;
         switch (mask) {
         case 0x00:
@@ -417,80 +524,67 @@ namespace SSE
         return _mm_or_si128(_mm_andnot_si128(c, a), _mm_and_si128(c, b));
     }
 
-    static Vc_INTRINSIC __m128i Vc_CONST mm_max_epi8 (__m128i a, __m128i b) {
-        return mm_blendv_epi8(b, a, _mm_cmpgt_epi8 (a, b));
+    Vc_INTRINSIC Vc_CONST __m128i max_epi8 (__m128i a, __m128i b) {
+        return blendv_epi8(b, a, _mm_cmpgt_epi8 (a, b));
     }
-    static Vc_INTRINSIC __m128i Vc_CONST mm_max_epi32(__m128i a, __m128i b) {
-        return mm_blendv_epi8(b, a, _mm_cmpgt_epi32(a, b));
+    Vc_INTRINSIC Vc_CONST __m128i max_epi32(__m128i a, __m128i b) {
+        return blendv_epi8(b, a, _mm_cmpgt_epi32(a, b));
     }
-//X         static Vc_INTRINSIC __m128i Vc_CONST mm_max_epu8 (__m128i a, __m128i b) {
-//X             return mm_blendv_epi8(b, a, _mm_cmpgt_epu8 (a, b));
+//X         Vc_INTRINSIC Vc_CONST __m128i max_epu8 (__m128i a, __m128i b) {
+//X             return _mm_blendv_epi8(b, a, cmpgt_epu8 (a, b));
 //X         }
-    static Vc_INTRINSIC __m128i Vc_CONST mm_max_epu16(__m128i a, __m128i b) {
-        return mm_blendv_epi8(b, a, _mm_cmpgt_epu16(a, b));
+    Vc_INTRINSIC Vc_CONST __m128i max_epu16(__m128i a, __m128i b) {
+        return blendv_epi8(b, a, cmpgt_epu16(a, b));
     }
-    static Vc_INTRINSIC __m128i Vc_CONST mm_max_epu32(__m128i a, __m128i b) {
-        return mm_blendv_epi8(b, a, _mm_cmpgt_epu32(a, b));
+    Vc_INTRINSIC Vc_CONST __m128i max_epu32(__m128i a, __m128i b) {
+        return blendv_epi8(b, a, cmpgt_epu32(a, b));
     }
-//X         static Vc_INTRINSIC __m128i Vc_CONST mm_min_epu8 (__m128i a, __m128i b) {
-//X             return mm_blendv_epi8(a, b, _mm_cmpgt_epu8 (a, b));
+//X         Vc_INTRINSIC Vc_CONST __m128i _mm_min_epu8 (__m128i a, __m128i b) {
+//X             return _mm_blendv_epi8(a, b, cmpgt_epu8 (a, b));
 //X         }
-    static Vc_INTRINSIC __m128i Vc_CONST mm_min_epu16(__m128i a, __m128i b) {
-        return mm_blendv_epi8(a, b, _mm_cmpgt_epu16(a, b));
+    Vc_INTRINSIC Vc_CONST __m128i min_epu16(__m128i a, __m128i b) {
+        return blendv_epi8(a, b, cmpgt_epu16(a, b));
     }
-    static Vc_INTRINSIC __m128i Vc_CONST mm_min_epu32(__m128i a, __m128i b) {
-        return mm_blendv_epi8(a, b, _mm_cmpgt_epu32(a, b));
+    Vc_INTRINSIC Vc_CONST __m128i min_epu32(__m128i a, __m128i b) {
+        return blendv_epi8(a, b, cmpgt_epu32(a, b));
     }
-    static Vc_INTRINSIC __m128i Vc_CONST mm_min_epi8 (__m128i a, __m128i b) {
-        return mm_blendv_epi8(a, b, _mm_cmpgt_epi8 (a, b));
+    Vc_INTRINSIC Vc_CONST __m128i min_epi8 (__m128i a, __m128i b) {
+        return blendv_epi8(a, b, _mm_cmpgt_epi8 (a, b));
     }
-    static Vc_INTRINSIC __m128i Vc_CONST mm_min_epi32(__m128i a, __m128i b) {
-        return mm_blendv_epi8(a, b, _mm_cmpgt_epi32(a, b));
+    Vc_INTRINSIC Vc_CONST __m128i min_epi32(__m128i a, __m128i b) {
+        return blendv_epi8(a, b, _mm_cmpgt_epi32(a, b));
     }
-    static Vc_INTRINSIC Vc_CONST __m128i mm_cvtepu8_epi16(__m128i epu8) {
+    Vc_INTRINSIC Vc_CONST __m128i cvtepu8_epi16(__m128i epu8) {
         return _mm_unpacklo_epi8(epu8, _mm_setzero_si128());
     }
-    static Vc_INTRINSIC Vc_CONST __m128i mm_cvtepi8_epi16(__m128i epi8) {
+    Vc_INTRINSIC Vc_CONST __m128i cvtepi8_epi16(__m128i epi8) {
         return _mm_unpacklo_epi8(epi8, _mm_cmplt_epi8(epi8, _mm_setzero_si128()));
     }
-    static Vc_INTRINSIC Vc_CONST __m128i mm_cvtepu16_epi32(__m128i epu16) {
+    Vc_INTRINSIC Vc_CONST __m128i cvtepu16_epi32(__m128i epu16) {
         return _mm_unpacklo_epi16(epu16, _mm_setzero_si128());
     }
-    static Vc_INTRINSIC Vc_CONST __m128i mm_cvtepi16_epi32(__m128i epu16) {
+    Vc_INTRINSIC Vc_CONST __m128i cvtepi16_epi32(__m128i epu16) {
         return _mm_unpacklo_epi16(epu16, _mm_cmplt_epi16(epu16, _mm_setzero_si128()));
     }
-    static Vc_INTRINSIC Vc_CONST __m128i mm_cvtepu8_epi32(__m128i epu8) {
-        return mm_cvtepu16_epi32(mm_cvtepu8_epi16(epu8));
+    Vc_INTRINSIC Vc_CONST __m128i cvtepu8_epi32(__m128i epu8) {
+        return cvtepu16_epi32(cvtepu8_epi16(epu8));
     }
-    static Vc_INTRINSIC Vc_CONST __m128i mm_cvtepi8_epi32(__m128i epi8) {
+    Vc_INTRINSIC Vc_CONST __m128i cvtepi8_epi32(__m128i epi8) {
         const __m128i neg = _mm_cmplt_epi8(epi8, _mm_setzero_si128());
         const __m128i epi16 = _mm_unpacklo_epi8(epi8, neg);
         return _mm_unpacklo_epi16(epi16, _mm_unpacklo_epi8(neg, neg));
     }
-    static Vc_INTRINSIC Vc_PURE __m128i mm_stream_load_si128(__m128i *mem) {
+    Vc_INTRINSIC Vc_PURE __m128i stream_load_si128(__m128i *mem) {
         return _mm_load_si128(mem);
     }
-
-} // namespace SSE
-} // namespace Vc
-/*OUTER_NAMESPACE_END*/
-#endif
-
-#ifdef VC_IMPL_POPCNT
-#include <popcntintrin.h>
+}  // namespace SseIntrinsics
+}  // namespace Vc
 #endif
 
 // SSE4.2
-#ifdef VC_IMPL_SSE4_2
-extern "C" {
-#include <nmmintrin.h>
-}
-#endif
-
-/*OUTER_NAMESPACE_BEGIN*/
-namespace Vc
+namespace Vc_VERSIONED_NAMESPACE
 {
-namespace SSE
+namespace SseIntrinsics
 {
     static Vc_INTRINSIC Vc_CONST float extract_float_imm(const __m128 v, const size_t i) {
         float f;
@@ -498,9 +592,9 @@ namespace SSE
         case 0:
             f = _mm_cvtss_f32(v);
             break;
-#if defined VC_IMPL_SSE4_1 && !defined VC_MSVC
+#if defined Vc_IMPL_SSE4_1 && !defined Vc_MSVC
         default:
-#ifdef VC_GCC
+#ifdef Vc_GCC
             f = __builtin_ia32_vec_ext_v4sf(static_cast<__v4sf>(v), (i));
 #else
             // MSVC fails to compile this because it can't optimize i to an immediate
@@ -528,7 +622,7 @@ namespace SSE
         return _mm_cvtsd_f64(_mm_castps_pd(_mm_movehl_ps(_mm_castpd_ps(v), _mm_castpd_ps(v))));
     }
     static Vc_INTRINSIC Vc_CONST float extract_float(const __m128 v, const size_t i) {
-#ifdef VC_GCC
+#ifdef Vc_GCC
         if (__builtin_constant_p(i)) {
             return extract_float_imm(v, i);
 //X         if (index <= 1) {
@@ -550,21 +644,21 @@ namespace SSE
     }
 
     static Vc_INTRINSIC Vc_PURE __m128  _mm_stream_load(const float *mem) {
-#ifdef VC_IMPL_SSE4_1
+#ifdef Vc_IMPL_SSE4_1
         return _mm_castsi128_ps(_mm_stream_load_si128(reinterpret_cast<__m128i *>(const_cast<float *>(mem))));
 #else
         return _mm_load_ps(mem);
 #endif
     }
     static Vc_INTRINSIC Vc_PURE __m128d _mm_stream_load(const double *mem) {
-#ifdef VC_IMPL_SSE4_1
+#ifdef Vc_IMPL_SSE4_1
         return _mm_castsi128_pd(_mm_stream_load_si128(reinterpret_cast<__m128i *>(const_cast<double *>(mem))));
 #else
         return _mm_load_pd(mem);
 #endif
     }
     static Vc_INTRINSIC Vc_PURE __m128i _mm_stream_load(const int *mem) {
-#ifdef VC_IMPL_SSE4_1
+#ifdef Vc_IMPL_SSE4_1
         return _mm_stream_load_si128(reinterpret_cast<__m128i *>(const_cast<int *>(mem)));
 #else
         return _mm_load_si128(reinterpret_cast<const __m128i *>(mem));
@@ -585,18 +679,70 @@ namespace SSE
     static Vc_INTRINSIC Vc_PURE __m128i _mm_stream_load(const unsigned char *mem) {
         return _mm_stream_load(reinterpret_cast<const int *>(mem));
     }
-} // namespace SSE
-} // namespace Vc
-/*OUTER_NAMESPACE_END*/
 
-// XOP / FMA4
-#if defined(VC_IMPL_XOP) || defined(VC_IMPL_FMA4)
-extern "C" {
-#include <x86intrin.h>
-}
+#ifndef __x86_64__
+    Vc_INTRINSIC Vc_PURE __m128i _mm_cvtsi64_si128(int64_t x) {
+        return _mm_castpd_si128(_mm_load_sd(reinterpret_cast<const double *>(&x)));
+    }
+#endif
+
+}  // namespace SseIntrinsics
+}  // namespace Vc
+
+namespace Vc_VERSIONED_NAMESPACE
+{
+namespace SSE
+{
+using namespace SseIntrinsics;
+
+template <typename T> struct ParameterHelper
+{
+    typedef T ByValue;
+    typedef T &Reference;
+    typedef const T &ConstRef;
+};
+
+template <typename T> struct VectorHelper
+{
+};
+
+template <typename T> struct VectorTypeHelper
+{
+    typedef __m128i Type;
+};
+template <> struct VectorTypeHelper<double>
+{
+    typedef __m128d Type;
+};
+template <> struct VectorTypeHelper<float>
+{
+    typedef __m128 Type;
+};
+
+template <typename T> struct DetermineGatherMask
+{
+    typedef T Type;
+};
+
+template <typename T> struct VectorTraits
+{
+    typedef typename VectorTypeHelper<T>::Type VectorType;
+    Vc_ALIGNED_TYPEDEF(sizeof(T), T, EntryType);
+    static constexpr size_t Size = sizeof(VectorType) / sizeof(EntryType);
+    enum Constants { HasVectorDivision = !std::is_integral<T>::value };
+    typedef Mask<T> MaskType;
+    typedef typename DetermineGatherMask<MaskType>::Type GatherMaskType;
+    typedef Common::VectorMemoryUnion<VectorType, EntryType> StorageType;
+};
+
+template <typename T> struct VectorHelperSize;
+}  // namespace SSE
+}  // namespace Vc
+
+#if defined(Vc_GCC) && !defined(__OPTIMIZE__)
+#pragma GCC diagnostic pop
 #endif
 
-#include "undomacros.h"
 #include "shuffle.h"
 
-#endif // SSE_INTRINSICS_H
+#endif // VC_SSE_INTRINSICS_H_
